/*
 * $Id: freebsd.c,v 1.8 1998/06/05 18:08:42 masaki Exp $
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/sysctl.h>
#include <syslog.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>

#include <sys/sockio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/route.h>
#include <net/if_dl.h>

#include <version.h>
#include <New.h>
#include <mrt.h>
#include <linked_list.h>
#include <interface.h>
#include <trace.h>
#include <rib.h>

#ifdef INRIA_IPV6
#include <netinet/in6_var.h>
#endif /* INRIA_IPV6 */


union sockunion {
    struct sockaddr sa;
    struct sockaddr_in sin;
    struct sockaddr_dl sdl;
#ifdef HAVE_IPV6
    struct sockaddr_in6 sin6;
#endif				/* HAVE_IPV6 */
};

static char *
sockunion2char (union sockunion *u)
{
    if (u->sa.sa_family == AF_INET) {
	return ((char *) &u->sin.sin_addr);
    }
#ifdef HAVE_IPV6
    else if (u->sa.sa_family == AF_INET6) {
	return ((char *) &u->sin6.sin6_addr);
    }
#endif /* HAVE_IPV6 */
    return (NULL);
}


int 
read_interfaces ()
{
    size_t needed;
    char *buf, *next, *lim;
    register struct if_msghdr *ifm;
    register struct ifa_msghdr *ifam;
    interface_t *interface;

    register int l;
    register char *cp;

    union sockunion so_dst, so_gate, so_mask, so_genmask, so_ifa, so_ifp,
      so_brd;

#define ROUNDUP(a) \
   ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))

#define NEXTADDRG(w, u) \
   if (ifm->ifm_addrs & (w)) {  \
           l = ROUNDUP(((struct sockaddr *)cp)->sa_len); \
           bcopy(cp, (char *)&(u), l); cp += l; \
   }

    int mib[] =
    {CTL_NET, PF_ROUTE, 0, 0, NET_RT_IFLIST, 0};

    if (sysctl (mib, 6, NULL, &needed, NULL, 0) < 0) {
	trace (TR_FATAL, INTERFACE_MASTER->trace, 
	       "sysctl: %s\n", strerror (errno));
	return (-1);
    }
    needed = ROUNDUP (needed);
    buf = NewArray (char, needed);
    if (sysctl (mib, 6, buf, &needed, NULL, 0) < 0) {
	trace (TR_FATAL, INTERFACE_MASTER->trace, 
	       "sysctl: %s\n", strerror (errno));
	return (-1);
    }

    lim = buf + needed;
    for (next = buf; next < lim; next += ifm->ifm_msglen) {

	int family = 0;
	char *addr = NULL;
	char *broadcast = NULL;
	int masklen = 0;

	ifm = (struct if_msghdr *) next;

	if (ifm->ifm_version != RTM_VERSION) {
	    trace (TR_ERROR, INTERFACE_MASTER->trace,
		   "ifm_version mismatch: %d should be %d\n",
		   ifm->ifm_version, RTM_VERSION);
	    return (-1);
	}

	if (ifm->ifm_type == RTM_IFINFO) {
            struct sockaddr_dl *sdl;
            u_char *cp;
            int len;
	    char name[IFNAMSIZ];

            cp = (u_char *)(ifm + 1);
            len = ifm->ifm_msglen;

            if (ifm->ifm_addrs != RTA_IFP) {
                trace (TR_ERROR, INTERFACE_MASTER->trace,
		       "bad ifm_addrs 0x%x\n", ifm->ifm_addrs);
                return (-1);
	    }
            sdl = (struct sockaddr_dl *) cp;
            if (sdl->sdl_family != AF_LINK) {
                trace (TR_ERROR, INTERFACE_MASTER->trace,
		       "bad sdl family %d\n", sdl->sdl_family);
                return (-1);
	    }
	    memcpy (name, sdl->sdl_data, sdl->sdl_nlen);
	    name[sdl->sdl_nlen] = '\0';
            if (sdl->sdl_nlen >= IFNAMSIZ) {
                trace (TR_ERROR, INTERFACE_MASTER->trace,
		       "too long name %d\n", sdl->sdl_nlen);
                return (-1);
            }

	    interface = new_interface (name, ifm->ifm_flags & 0xffff, 
				       ifm->ifm_data.ifi_mtu, 
				       ifm->ifm_index);
	}
	if (ifm->ifm_type == RTM_NEWADDR) {
	    ifam = (struct ifa_msghdr *) next;
	    cp = (char *) ifam + sizeof (*ifam);

	    NEXTADDRG (RTA_DST, so_dst);
	    NEXTADDRG (RTA_GATEWAY, so_gate);
	    NEXTADDRG (RTA_NETMASK, so_mask);
	    NEXTADDRG (RTA_GENMASK, so_genmask);
	    NEXTADDRG (RTA_IFP, so_ifp);
	    NEXTADDRG (RTA_IFA, so_ifa);
	    NEXTADDRG (RTA_BRD, so_brd);

	    if ((interface = find_interface_byindex (ifam->ifam_index)) == NULL)
	        continue;

	    if (ifam->ifam_addrs & RTA_BRD) {
		broadcast = sockunion2char (&so_brd);
	    }
	    if (ifam->ifam_addrs & RTA_IFA) {
		family = so_ifa.sa.sa_family;
		addr = sockunion2char (&so_ifa);
#ifdef WIDE_IPV6
		/* KAME IPV6 still has an index here */
		if (family == AF_INET6 && 
			IN6_IS_ADDR_LINKLOCAL ((struct in6_addr *)addr)) {
            	    addr[2] = 0;
            	    addr[3] = 0;
		}
#endif /* WIDE_IPV6 */
		if (ifam->ifam_addrs & RTA_NETMASK) {
		    so_mask.sa.sa_family = so_ifa.sa.sa_family;
		    masklen = mask2len (sockunion2char (&so_mask),
					(family == AF_INET) ? 4 :
#ifdef HAVE_IPV6
					(family == AF_INET6) ? 16 :
#endif /* HAVE_IPV6 */
					0);
		}
		if (addr)
		    add_addr_to_interface (interface, family, addr, masklen,
				       (broadcast) ? broadcast : 0);
	    }
	}
    }
    Delete (buf);
    return (1);
}


struct m_rtmsg {
    struct rt_msghdr m_rtm;
    char m_space[512];
};


static int route_sockfd = -1;

int
kernel_init ()
{
    if ((route_sockfd = socket (PF_ROUTE, SOCK_RAW, 0)) < 0) {
        trace (TR_ERROR, MRT->trace, "PF_ROUTE socket (%s)\n", 
	       strerror (errno));
        return (-1);
    }   
    return (0);
}


/* 
 * kernel_update_route
 *   cmd should be one of:
 *     KERNEL_ROUTE_ADD (0)
 *     KERNEL_ROUTE_CHG (1)
 *     KERNEL_ROUTE_DEL (2)
 * some flags for the route should be passed
 */
int 
kernel_update_route (int cmd, prefix_t * dest, prefix_t * nexthop, 
			prefix_t *oldhop, int index, int oldindex)
{
    int rlen;
    struct m_rtmsg m_rtmsg;
    char *cp = m_rtmsg.m_space;
    union sockunion so_dst, so_gate, so_mask, so_ifa, so_ifp;

    int flags, rtm_addrs;
    static int seq = 0;
    struct rt_metrics rt_metrics;
    register int l;
#ifdef notdef
    interface_t *interface;
#endif
    int len = dest->bitlen;

    if (route_sockfd < 0)
	return (-1);

    bzero ((char *) &m_rtmsg, sizeof (m_rtmsg));
    bzero ((char *) &rt_metrics, sizeof (rt_metrics));
    bzero (&so_dst, sizeof (so_dst));
    bzero (&so_gate, sizeof (so_gate));
    bzero (&so_mask, sizeof (so_mask));
    bzero (&so_ifp, sizeof (so_ifp));
    bzero (&so_ifa, sizeof (so_ifa));

    if (dest->family == AF_INET) {
	so_dst.sin.sin_addr.s_addr = prefix_tolong (dest);
	so_dst.sin.sin_len = sizeof (struct sockaddr_in);
	so_dst.sin.sin_family = AF_INET;
    	netmasking (AF_INET, (char *)&so_dst.sin.sin_addr, len);

	so_gate.sin.sin_addr.s_addr = prefix_tolong (nexthop);
	so_gate.sin.sin_len = sizeof (struct sockaddr_in);
	so_gate.sin.sin_family = AF_INET;

	so_mask.sin.sin_len = sizeof (struct sockaddr_in);
        so_mask.sin.sin_family = AF_UNSPEC;
	len2mask (len, (char *) &so_mask.sin.sin_addr, 4);

	rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK;

	if (index > 0) {
	    so_ifp.sdl.sdl_index = index;
	    so_ifp.sdl.sdl_len = sizeof (struct sockaddr_dl);
	    so_ifp.sdl.sdl_family = AF_LINK;
	    rtm_addrs |= RTA_IFP;
	}


    }
#ifdef HAVE_IPV6
    else if (dest->family == AF_INET6) {

	memcpy (&so_dst.sin6.sin6_addr, prefix_tochar (dest), 16);
	so_dst.sin6.sin6_len = sizeof (struct sockaddr_in6);
	so_dst.sin6.sin6_family = AF_INET6;
    	netmasking (AF_INET6, (char *)&so_dst.sin6.sin6_addr, len);

        memcpy (&so_gate.sin6.sin6_addr, prefix_tochar (nexthop), 16);
	so_gate.sin6.sin6_len = sizeof (struct sockaddr_in6);
	so_gate.sin6.sin6_family = AF_INET6;

	so_mask.sin6.sin6_len = sizeof (struct sockaddr_in6);
        so_mask.sin6.sin6_family = AF_UNSPEC;
	len2mask (len, (char *) &so_mask.sin6.sin6_addr, 16);

	rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK;

	if (index > 0) {
	     so_ifp.sdl.sdl_index = index;
	     so_ifp.sdl.sdl_len = sizeof (struct sockaddr_dl);
	     so_ifp.sdl.sdl_family = AF_LINK;
	     rtm_addrs |= RTA_IFP;
	}

#ifdef WIDE_IPV6
	/* KAME IPV6 still requires an index here */
	if (IN6_IS_ADDR_LINKLOCAL (&so_gate.sin6.sin6_addr)) {
       	    so_gate.sin6.sin6_addr.s6_addr8[2] = index >> 8;;
       	    so_gate.sin6.sin6_addr.s6_addr8[3] = index;
	}
#endif /* WIDE_IPV6 */
#ifdef notdef
	interface = find_interface_byindex (index);

	assert (interface->primary6);
	memcpy (&so_ifa.sin6.sin6_addr,
		prefix_tochar (
			BIT_TEST (interface->flags, IFF_POINTOPOINT)?
			    interface->primary6->broadcast:
			interface->primary6->prefix), 16);
	so_ifa.sin6.sin6_len = sizeof (struct sockaddr_in6);
	so_ifa.sin6.sin6_family = AF_INET6;
#endif

    }
#endif /* HAVE_IPV6 */
    else			/* unknown family */
	return (-1);

    flags = RTF_UP;
    if (dest->family == AF_INET) {
#ifdef notdef
	if (len == 32)
	    flags |= RTF_HOST;
#endif
	if (so_dst.sin.sin_addr.s_addr != INADDR_ANY)
	    flags |= RTF_GATEWAY;
    }
#ifdef HAVE_IPV6
    else if (dest->family == AF_INET6) {
#ifdef notdef
	if (len == 128)
	    flags |= RTF_HOST;
#endif
	if (!ipv6_any_addr (&so_dst.sin6.sin6_addr))
	    flags |= RTF_GATEWAY;
    }
#endif /* HAVE_IPV6 */

#define NEXTADDRP(w, u) \
	if (rtm_addrs & (w)) {\
	    l = ROUNDUP(u.sa.sa_len); bcopy((char *)&(u), cp, l); cp += l;\
	}

    assert (cmd == KERNEL_ROUTE_ADD || cmd == KERNEL_ROUTE_DEL ||
            cmd == KERNEL_ROUTE_CHG);
    if (cmd == KERNEL_ROUTE_ADD)
	m_rtmsg.m_rtm.rtm_type = RTM_ADD;
    else if (cmd == KERNEL_ROUTE_DEL)
	m_rtmsg.m_rtm.rtm_type = RTM_DELETE;
    else if (cmd == KERNEL_ROUTE_CHG)
	m_rtmsg.m_rtm.rtm_type = RTM_CHANGE;

    m_rtmsg.m_rtm.rtm_flags = flags;
    m_rtmsg.m_rtm.rtm_version = RTM_VERSION;
    m_rtmsg.m_rtm.rtm_seq = ++seq;
    m_rtmsg.m_rtm.rtm_addrs = rtm_addrs;
    m_rtmsg.m_rtm.rtm_rmx = rt_metrics;
    /*m_rtmsg.m_rtm.rtm_inits = 0; */

    NEXTADDRP (RTA_DST, so_dst);
    NEXTADDRP (RTA_GATEWAY, so_gate);
    NEXTADDRP (RTA_NETMASK, so_mask);
    NEXTADDRP (RTA_IFP, so_ifp);
    NEXTADDRP (RTA_IFA, so_ifa);

    m_rtmsg.m_rtm.rtm_msglen = l = cp - (char *) &m_rtmsg;

    if ((rlen = write (route_sockfd, (char *) &m_rtmsg, l)) < 0) {
	if (cmd == KERNEL_ROUTE_CHG) {
	    int r = 0;
	    trace (TR_WARN, MRT->trace, 
		   "route socket dst %s nh %s (%s), trying delete/add\n", 
	            prefix_toax (dest), prefix_toa (nexthop), strerror (errno));
	    r = kernel_update_route (KERNEL_ROUTE_DEL, dest, oldhop, 
					NULL, oldindex, 0);
	    r = kernel_update_route (KERNEL_ROUTE_ADD, dest, nexthop, 
					NULL, index, 0);
	    return (r);
	}
	if (errno == EEXIST) {
	    return (1);
	}			/* route already exists */
	trace (TR_ERROR, MRT->trace, "route socket dst %s nh %s (%s)\n", 
	       prefix_toax (dest), prefix_toa (nexthop), strerror (errno));
	return (-1);
    }
    return (1);
}


#define NEXTADDRR(w, u) \
   if (rtm->rtm_addrs & (w)) {  \
           l = ROUNDUP(((struct sockaddr *)cp)->sa_len); \
           bcopy(cp, (char *)&(u), l); cp += l; \
   }

int
kernel_read_rt_table ()
{

    union sockunion so_dst, so_gate, so_mask, so_genmask, so_ifa, so_ifp;
    size_t needed;
    int mib[] =
    {CTL_NET, PF_ROUTE, 0, 0, NET_RT_DUMP, 0};
    char *buf, *next, *lim, *cp;
    struct rt_msghdr *rtm;

    int family, masklen, l, index = 0;
    u_char *dest, *nexthop, *mask;

    if (sysctl (mib, 6, NULL, &needed, NULL, 0) < 0) {
	trace (TR_ERROR, MRT->trace, "sysctl: %s\n", strerror (errno));
	return (-1);
    }
    buf = NewArray (char, needed);
    if (sysctl (mib, 6, buf, &needed, NULL, 0) < 0) {
	trace (TR_ERROR, MRT->trace, "sysctl: %s\n", strerror (errno));
	return (-1);
    }

    lim = buf + needed;
    for (next = buf; next < lim; next += rtm->rtm_msglen) {
	rtm = (struct rt_msghdr *) next;

	if (rtm->rtm_version != RTM_VERSION) {
	    trace (TR_ERROR, MRT->trace,
		   "rtm_version mismatch: %d should be %d\n",
		   rtm->rtm_version, RTM_VERSION);
	}

	if (rtm->rtm_type == RTM_GET) {

#if 0
	    printf ("rtm->rtm_flags=%x rtm->rtm_addrs=%x\n", 
		rtm->rtm_flags, rtm->rtm_addrs);
#endif
	    /* RTF_PRCLONING is set on KAME IPV6 */
	    /* REJECT and BLACKHOLE are not handled yet */
	    if (BIT_TEST (rtm->rtm_flags, 
		    ~(RTF_UP|RTF_GATEWAY|RTF_HOST|RTF_STATIC|
			RTF_REJECT|RTF_BLACKHOLE|RTF_PRCLONING)))
		continue;
	    if (!BIT_TEST (rtm->rtm_flags, RTF_UP))
		continue;
	    if (!BIT_TEST (rtm->rtm_flags, RTF_GATEWAY))
		continue;

	    index = rtm->rtm_index;
	    cp = (char *) rtm + sizeof (*rtm);

	    NEXTADDRR (RTA_DST, so_dst);
	    NEXTADDRR (RTA_GATEWAY, so_gate);
	    NEXTADDRR (RTA_NETMASK, so_mask);
	    NEXTADDRR (RTA_GENMASK, so_genmask);
	    NEXTADDRR (RTA_IFP, so_ifp);
	    NEXTADDRR (RTA_IFA, so_ifa);

	    family = so_dst.sa.sa_family;
	    dest = sockunion2char (&so_dst);
	    nexthop = sockunion2char (&so_gate);
	    so_mask.sa.sa_family = family; /* required by sockunion2char */
	    mask = sockunion2char (&so_mask);

	    if (family != AF_INET
#ifdef HAVE_IPV6
		&& family != AF_INET6
#endif /* HAVE_IPV6 */
		)
		continue;

	    if (BIT_TEST (rtm->rtm_flags, RTF_HOST)) {
		masklen = (family == AF_INET) ? 32 : 128;
	    }
	    else {
		/* I don't know the reason, 
		   but it's needed for getting rid of strange subnetmask */
		if (family == AF_INET && *(u_long *) dest == *(u_long *) mask)
		    so_mask.sa.sa_len = 0;

		if (!BIT_TEST (rtm->rtm_addrs, RTA_NETMASK) || 
			so_mask.sa.sa_len <= 2) {
		    /* not specific netmask */
		    if (family == AF_INET) {
			/* natural mask */
			if (*(u_long *) dest == INADDR_ANY)
			    masklen = 0;
			else
			    masklen = (dest[0] < 128) ? 8 :
				(dest[0] < 192) ? 16 :
				(dest[0] < 224) ? 24 : 32;
		    }
		    else {
			masklen = 0;
		    }
		}
		else {
		    char tmp[16];
		    memset (tmp, 0, sizeof (tmp));
		    memcpy (tmp, sockunion2char (&so_mask), so_mask.sa.sa_len -
			    (sockunion2char (&so_mask) - (char *) &so_mask));
		    masklen = mask2len (tmp, (family == AF_INET) ? 4 : 16);
		}
	    }

	    if (so_gate.sa.sa_family != family)
		continue;

#ifdef WIDE_IPV6
	    /* KAME IPV6 still has an index here */
	    if (family == AF_INET6 && 
		    IN6_IS_ADDR_LINKLOCAL ((struct in6_addr *)nexthop)) {
       	        nexthop[2] = 0;
       	        nexthop[3] = 0;
	    }
#endif /* WIDE_IPV6 */

	    add_kernel_route (family, dest, nexthop, masklen, index);

	}
    }
    Delete (buf);
    return (1);
}
